import tempfile

import os
import pytest
import time

from twindb_infrastructure.providers.aws import launch_ec2_instance, terminate_instance, AwsError, get_instance_state, \
    ec2_describe_instance, add_name_tag, start_instance, get_instance_private_ip, get_instance_public_ip, stop_instance

AWS_INSTANCE_STATES = [
    'running',
    'shutting-down',
    'pending',
    'stopping',
    'stopped',
    'terminated',
    'rebooting',
]


@pytest.fixture
def instance_id(request):
    instance_profile = {
        "ImageId": "ami-cd0f5cb6",
        "InstanceType": "t2.micro",
        "KeyName": "travis_ci",
        "SecurityGroupIds": ["sg-c63071b8"],
        "SubnetId": "subnet-07c5404f",
        "RootVolumeSize": 200,
        "InstanceInitiatedShutdownBehavior": 'stop',
        "AvailabilityZone": "us-east-1c",
        "UserName": "ubuntu",
        "Name": "integraion-test",
        "Region": "us-east-1"
    }
    inst_id = launch_ec2_instance(instance_profile, region=instance_profile['Region'], private_key_file="~/.ssh/id_rsa")
    assert get_instance_state(inst_id) == "running"

    def resource_teardown():
        terminate_instance(inst_id)
        time.sleep(2)
        assert get_instance_state(inst_id) == "shutting-down"

    request.addfinalizer(resource_teardown)
    return inst_id


def test_instance_state(instance_id):
    assert get_instance_state(instance_id) in AWS_INSTANCE_STATES


def test_describe_instance(instance_id):
    assert ec2_describe_instance(instance_id)


def test_add_name_tag(instance_id):
    assert add_name_tag(instance_id, 'Test name')
    response = ec2_describe_instance(instance_id)
    assert response["Reservations"][0]["Instances"][0]['Tags'][0]['Value'] == 'Test name'


def test_get_instance_private_ip(instance_id):
    assert get_instance_private_ip(instance_id)


def test_get_instance_public_ip(instance_id):
    assert get_instance_public_ip(instance_id)


def test_stop_start_instance(instance_id):
    stop_instance(instance_id)
    flag = False
    time_over = time.time() + 15 * 60
    while time.time() < time_over:
        if get_instance_state(instance_id) == "stopped":
            flag = True
            break
        time.sleep(1)
    assert flag
    start_instance(instance_id)
    flag = False
    time_over = time.time() + 15 * 60

    while time.time() < time_over:
        if get_instance_state(instance_id) == "running":
            flag = True
            break
        time.sleep(1)
    assert flag





import pytest

from twindb_infrastructure.check_http import CheckHttpResponse


@pytest.mark.parametrize('http_code, expected_response', [
    (
        503,
        """HTTP/1.1 503 Service unavailable
Content-Type: text/plain; charset=UTF-8
Connection: close
Content-Length: 4

foo
"""
    ),
    (
        200,
        """HTTP/1.1 200 OK
Content-Type: text/plain; charset=UTF-8
Connection: close
Content-Length: 4

foo
"""
    )
])
def test_str(http_code, expected_response):
    resp = CheckHttpResponse(
        message='foo',
        http_code=http_code,
        nagios_code=4
    )
    assert str(resp) == expected_response



from contextlib import contextmanager

import boto3
from subprocess import check_call, CalledProcessError

import pymysql
from pymysql import MySQLError
from pymysql.cursors import DictCursor

from twindb_infrastructure import log
from twindb_infrastructure.util import domainname


def start_proxy(proxy):
    """Start ProxySQL on a remote server proxy"""
    cmd = [
        'sudo',
        'ssh',
        '-t',
        proxy,
        'sudo /etc/init.d/proxysql start'
    ]
    log.info('Executing: %s', ' '.join(cmd))
    check_call(cmd)


def stop_proxy(proxy):
    """Stop ProxySQL on a remote server proxy"""
    cmd = [
        'sudo',
        'ssh',
        '-t',
        proxy,
        'sudo killall -9 proxysql'
    ]
    log.info('Executing: %s', ' '.join(cmd))
    check_call(cmd)


def restart_proxy(proxy):
    """Restart ProxySQL on server proxy"""
    log.info('Restarting %s', proxy)
    stop_proxy(proxy)
    start_proxy(proxy)


def change_names_to(names, ip_addr):
    client = boto3.client('route53')
    if names:
        for name in names:
            log.info('Updating A record of %s to %s', name, ip_addr)

            print(name)
            response = client.list_hosted_zones_by_name(
                DNSName=domainname(name),
            )
            zone_id = response['HostedZones'][0]['Id']
            request = {
                'HostedZoneId': zone_id,
                'ChangeBatch': {
                    'Comment': 'Automated switchover DNS update',
                    'Changes': [
                        {
                            'Action': 'UPSERT',
                            'ResourceRecordSet': {
                                'Name': name,
                                'Type': 'A',
                                'TTL': 300,
                                'ResourceRecords': [
                                    {
                                        'Value': ip_addr
                                    },
                                ]
                            }
                        }
                    ]
                }
            }
            client.change_resource_record_sets(**request)


def log_remaining_sessions(host, user='root', password='', port=3306):
    """Connect to host and print existing sessions.

    :return: Number of connected sessions
    :rtype: int
    """
    with _connect(host, user=user, password=password, port=port) as conn:
        cursor = conn.cursor()
        query = "SHOW PROCESSLIST"
        cursor.execute(query)
        nrows = cursor.rowcount
        while True:
            row = cursor.fetchone()
            if row:
                print(row)
            else:
                break

    return nrows


def server_ready(host, user='root', password='', port=3306):
    """Connect to host and execute SELECT 1 to make sure
    it's up and running.

    :return: True if server is ready for connections
    :rtype: bool
    """
    try:
        with _connect(host, user=user, password=password, port=port) as conn:
            cursor = conn.cursor()
            query = "SELECT 1"
            cursor.execute(query)
            return True
    except MySQLError as err:
        log.error(err)
        return False


def eth1_present(proxy):
    """Check if eth1 is up on remote host proxy"""
    cmd = [
        'sudo',
        'ssh',
        '-t',
        proxy,
        '/sbin/ifconfig eth1'
    ]
    log.info('Executing: %s', ' '.join(cmd))
    try:
        check_call(cmd)
        return True
    except CalledProcessError:
        return False


@contextmanager
def _connect(host, user='root', password='', port=3306):
    """Connect to ProxySQL admin interface."""
    pass
    connect_args = {
        'host': host,
        'port': port,
        'user': user,
        'passwd': password,
        'connect_timeout': 60,
        'cursorclass': DictCursor
    }

    conn = pymysql.connect(**connect_args)
    yield conn
    conn.close()



import re
import socket
from time import sleep

from requests import RequestException

NAGIOS_EXIT_OK = 0
NAGIOS_EXIT_WARNING = 1
NAGIOS_EXIT_CRITICAL = 2
NAGIOS_EXIT_UNKNOWN = 3


class CheckResponse(object):
    def __init__(self, **kwargs):
        self._message = kwargs.get('message', 'Service unavailable')
        self._nagios_code = kwargs.get('nagios_code', NAGIOS_EXIT_UNKNOWN)

    @property
    def nagios_code(self):
        return self._nagios_code

    def __str__(self):
        return self._message


class CheckHttpResponse(CheckResponse):
    def __init__(self, **kwargs):
        super(CheckHttpResponse, self).__init__(**kwargs)
        self._http_code = kwargs.get('http_code', 503)

    @property
    def http_code(self):
        return self._http_code

    def __str__(self):
        return """HTTP/1.1 {code} {short_message}
Content-Type: text/plain; charset=UTF-8
Connection: close
Content-Length: {len}

{msg}
""".format(
            code=self._http_code,
            short_message="OK" if self._http_code == 200
            else "Service unavailable",
            len=len(self._message) + 1,
            msg=self._message
        )


class HttpChecker(object):
    __attributes = [
        'critical_load_time',
        'warning_load_time',
        'title',
        'title_regexp',
        'body_regexp'
    ]

    def __init__(self, **kwargs):
        self._body_regexp = None
        self._title_regexp = None
        self._title = None
        self._critical_load_time = None
        self._warning_load_time = None

        for attr in self.__attributes:
            setattr(
                self,
                '_%s' % attr,
                kwargs.get(attr, None)
            )

    def check(self, loader, resp_class):
        """

        :param loader:
        :type loader: Loader
        :param resp_class: What response type the method should return
        :type resp_class: class
        :return: Response
        :rtype: CheckResponse
        """
        try:
            # if load time more than critical threshold
            if self._critical_load_time \
                    and loader.load_time > self._critical_load_time:

                kwargs = {
                    'message': 'CRITICAL - %s: load time %f '
                               'seconds more than %f'
                               % (
                                   loader.url,
                                   loader.load_time,
                                   self._critical_load_time
                               ),
                    'nagios_code': NAGIOS_EXIT_CRITICAL
                }
                if resp_class == CheckHttpResponse:
                    kwargs['http_code'] = 503

                return resp_class(**kwargs)

            # if load time more than warning threshold
            if self._warning_load_time \
                    and loader.load_time > self._warning_load_time:

                kwargs = {
                    'message': 'WARNING - %s: load time %f '
                               'seconds more than %f'
                               % (
                                   loader.url,
                                   loader.load_time,
                                   self._warning_load_time
                               ),
                    'nagios_code': NAGIOS_EXIT_WARNING
                }
                if resp_class == CheckHttpResponse:
                    kwargs['http_code'] = 200

                return resp_class(**kwargs)

            # If title is different than expected
            if self._title and self._title != loader.title:
                kwargs = {
                    'message': "CRITICAL - %s: "
                               "Expected title %s. Actual title '%s'"
                               % (loader.url, self._title, loader.title),
                    'nagios_code': NAGIOS_EXIT_CRITICAL
                }
                if resp_class == CheckHttpResponse:
                    kwargs['http_code'] = 503

                return resp_class(**kwargs)

            # If title doesn't match expected regexp
            if self._title_regexp \
                    and re.match(self._title_regexp, loader.title) is None:

                kwargs = {
                    'message': "CRITICAL - %s: Title '%s' "
                               "is expected to match regexp '%s'"
                               % (loader.url, loader.title, self._title_regexp),
                    'nagios_code': NAGIOS_EXIT_CRITICAL
                }
                if resp_class == CheckHttpResponse:
                    kwargs['http_code'] = 503

                return resp_class(**kwargs)

            # If body doesn't match expected regexp
            if self._body_regexp \
                    and re.match(self._body_regexp, loader.body) is None:

                kwargs = {
                    'message': "CRITICAL - %s: Body '%s...' "
                               "is expected to match regexp '%s'"
                               % (
                                   loader.url,
                                   loader.body[0:16],
                                   self._body_regexp
                               ),
                    'nagios_code': NAGIOS_EXIT_CRITICAL
                }
                if resp_class == CheckHttpResponse:
                    kwargs['http_code'] = 503

                return resp_class(**kwargs)

            # If no checks fails respond with success
            loader.load()
            kwargs = {
                'message': "OK - %s is healthy"
                           % (
                               loader.url,
                           ),
                'nagios_code': NAGIOS_EXIT_OK
            }
            if resp_class == CheckHttpResponse:
                kwargs['http_code'] = 200

        except RequestException as err:
            kwargs = {
                'message': "CRITICAL - {url}: {err_msg}".format(
                    url=loader.url,
                    err_msg=err
                ),
                'nagios_code': NAGIOS_EXIT_CRITICAL
            }

            if resp_class == CheckHttpResponse:
                kwargs['http_code'] = 503

        return resp_class(**kwargs)

    def start_server(self, http_port, loader):
        try:
            s = socket.socket()

            while True:
                try:
                    s.bind(('', http_port))
                    break
                except socket.error as error:
                    # if socket in TIME_WAIT, sleep a bit and retry
                    if error.errno in [48, 98]:
                        sleep(5)
                    else:
                        raise
            s.listen(1)

            while True:
                c, addr = s.accept()
                response = self.check(
                    loader,
                    CheckHttpResponse
                )
                loader.reset()
                c.sendall(str(response))
                # Receive all outstanding bytes, otherwise the client will
                # get a RST
                c.recv(4096)
                c.close()

        except KeyboardInterrupt:
            return



"""twindb-monitoring CLI module."""
import click

from twindb_infrastructure.check_http import \
    HttpChecker, CheckHttpResponse, CheckResponse
from twindb_infrastructure.loader import Loader


@click.group()
@click.version_option()
def main():
    pass


@main.command()
@click.argument('url')
@click.option(
    '-w', '--warning',
    help='Response time to result in warning status (seconds)',
    type=click.FLOAT,
)
@click.option(
    '-c', '--critical',
    help='Response time to result in critical status (seconds)',
    type=click.FLOAT,
)
@click.option(
    '-t', '--timeout',
    help='Seconds before connection times out',
    default=10,
    show_default=True,
    type=click.INT,
)
@click.option(
    '--title',
    help='Expect the string to be a title',
)
@click.option(
    '--title-regexp',
    help='Expect the regexp to match the title',
)
@click.option(
    '--body-regexp',
    help='Expect the regexp to match the body',
)
@click.option(
    '--http',
    help='Print result as an HTTP response',
    is_flag=True,
    default=False
)
@click.option(
    '--host',
    help='Pass this value as Host: HTTP header',
)
@click.option(
    '--protocol',
    help='This value is used for X-Forwarded-Proto header',
    default='http',
    show_default=True,
)
@click.option(
    '--http-server',
    help='Run an HTTP server that reports the check result',
    is_flag=True,
    default=False
)
@click.option(
    '--http-port',
    help='Bind the HTTP server to this TCP port',
    type=click.INT,
    default=8080,
    show_default=True,
)
def check_http(url,
               warning,
               critical,
               timeout,
               title,
               title_regexp,
               body_regexp,
               http,
               host,
               protocol,
               http_server,
               http_port
               ):
    """
    Make an HTTP(s) GET request and check response against given criteria.

    The exit code matches Nagios convention:

    \b
        - 0 - OK
        - 1 - Warning
        - 2 - Critical
        - 3 - Unknown
    """
    loader = Loader(
        url,
        timeout=timeout,
        host=host,
        protocol=protocol
    )
    checker = HttpChecker(
        critical_load_time=critical,
        warning_load_time=warning,
        title=title,
        title_regexp=title_regexp,
        body_regexp=body_regexp
    )

    if http_server:
        checker.start_server(http_port, loader)

    else:
        response = checker.check(
            loader,
            CheckHttpResponse if http else CheckResponse
        )
        print(response)
        exit(response.nagios_code)



import mock
import pytest
from requests import RequestException

from twindb_infrastructure.check_http import HttpChecker, CheckHttpResponse, \
    NAGIOS_EXIT_CRITICAL, NAGIOS_EXIT_OK
from twindb_infrastructure.loader import Loader


def test_check_critical():
    ch = HttpChecker(
        critical_load_time=1
    )
    mock_loader = mock.Mock()
    mock_loader.load_time = 2.0
    resp = ch.check(
        mock_loader,
        CheckHttpResponse
    )
    assert resp.nagios_code == NAGIOS_EXIT_CRITICAL
    assert resp.http_code == 503


@pytest.mark.parametrize(
    'actual_title, title_regexp, nagios_code, http_code',
    [
        (
            'foo', 'foo', NAGIOS_EXIT_OK, 200
        ),
        (
            'foo bar', 'foo$', NAGIOS_EXIT_CRITICAL, 503
        )
    ]
)
def test_check_title_regexp(actual_title, title_regexp, nagios_code, http_code):
    ch = HttpChecker(
        title_regexp=title_regexp
    )
    mock_loader = mock.Mock()
    mock_loader.title = actual_title
    resp = ch.check(
        mock_loader,
        CheckHttpResponse
    )
    assert resp.nagios_code == nagios_code
    assert resp.http_code == http_code


@mock.patch.object(Loader, '_get_tag')
def test_check_on_exception(mock_get_tag):
    ch = HttpChecker(
        title_regexp='foo'
    )
    mock_get_tag.side_effect = RequestException
    resp = ch.check(
        Loader('foo-url'),
        CheckHttpResponse
    )
    assert resp.nagios_code == NAGIOS_EXIT_CRITICAL
    assert resp.http_code == 503



import shlex
import subprocess
import time
from twindb_infrastructure import log


def run_command(command):
    log.debug("Executing: %r" % command)
    try:
        process = subprocess.Popen(shlex.split(command),
                                   stdout=subprocess.PIPE)
        while True:
            output = process.stdout.readline()
            if output == '' and process.poll() is not None:
                break
            if output:
                log.info(output)
        rc = process.poll()
        log.debug("Exit code: %d" % rc)
        if rc != 0:
            raise subprocess.CalledProcessError(rc, command)
        return rc
    except (OSError, ValueError) as err:
        log.error(err)
        return -1


def disable_selinux(ip, username, key):

    try:
        cmd_common = ["ssh", "-t",
                      "-o", "StrictHostKeyChecking=no",
                      "-i", key,
                      "-l", username,
                      ip,
                      "sudo"]
        cmd = cmd_common + ["sed", "-i",
                            "s/SELINUX=enforcing/SELINUX=disabled/",
                            "/etc/selinux/config", "/etc/sysconfig/selinux"]
        log.debug("Executing: %r" % cmd)
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)

        try:
            cmd = cmd_common + ["reboot"]
            log.debug("Executing: %r" % cmd)
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            pass

        if not wait_sshd(ip, key, username):
            return False
        return True

    except subprocess.CalledProcessError as err:
        log.error(err.output)
        return False


def wait_sshd(ip, key_file, username):
    # Wait will sshd is up
    timeout = 300
    log.info("Waiting till sshd on %s starts" % ip)
    while subprocess.call(["ssh", "-o", "StrictHostKeyChecking=no",
                           "-o", "PasswordAuthentication=no",
                           "-i", key_file, "-l", username, ip,
                           "ls"]) != 0:
        log.info("Waiting %d more seconds" % timeout)
        time.sleep(3)
        timeout -= 3
        if timeout == 0:
            log.error("Failed to start sshd. Timeout expired")
            return False
    return True



from boto3.exceptions import ResourceNotExistsError, UnknownAPIVersionError
from botocore.exceptions import ClientError
from click.testing import CliRunner
import mock
import pytest
from twindb_infrastructure import twindb_aws
from twindb_infrastructure.config.config import Config, ConfigException
from twindb_infrastructure.providers.aws import start_instance, terminate_instance, stop_instance, ec2_describe_instance, \
    AwsError, get_instance_state, get_instance_private_ip, get_instance_public_ip, add_name_tag, associate_address
from twindb_infrastructure.twindb_aws import parse_config


@mock.patch.object(Config, '__init__')
def test_parse_config(mock_config):
    mock_config.return_value = None
    parse_config('foo')


@mock.patch.object(Config, '__init__')
def test_parse_config_exception(mock_config):
    mock_config.return_value = None
    mock_config.side_effect = ConfigException
    with pytest.raises(ConfigException):
        parse_config('foo')


def test_main():
    runner = CliRunner()
    result = runner.invoke(twindb_aws.main)
    assert result.exit_code == 0


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_start_instance(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_instance = mock.Mock()
    mock_ec2.instances.filter.return_value = mock_instance
    start_instance('foo-bar')
    mock_boto3.resource.assert_called_once_with('ec2')
    mock_ec2.instances.filter.assert_called_once_with(InstanceIds=['foo-bar'])
    mock_instance.start.assert_called_once()


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_terminate_instance(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_instance = mock.Mock()
    mock_ec2.instances.filter.return_value = mock_instance
    terminate_instance('foo-bar')
    mock_boto3.resource.assert_called_once_with('ec2')
    mock_ec2.instances.filter.assert_called_once_with(InstanceIds=['foo-bar'])
    mock_instance.terminate.assert_called_once()


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_stop_instance(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_instance = mock.Mock()
    mock_ec2.instances.filter.return_value = mock_instance
    stop_instance('foo-bar')
    mock_boto3.resource.assert_called_once_with('ec2')
    mock_ec2.instances.filter.assert_called_once_with(InstanceIds=['foo-bar'])
    mock_instance.stop.assert_called_once()



@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_start_instance_resource_exception(mock_boto3):
    mock_boto3.resource.side_effect = ResourceNotExistsError(mock.Mock(), 'error', '')
    with pytest.raises(AwsError):
        start_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_start_instance_api_exception(mock_boto3):
    mock_boto3.resource.side_effect = UnknownAPIVersionError(mock.Mock(), 'error', '')
    with pytest.raises(AwsError):
        start_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_start_instance_exception(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_instance = mock.Mock()
    mock_ec2.instances.filter.return_value = mock_instance
    mock_instance.start.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])
    with pytest.raises(AwsError):
        start_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_start_instance_filter_exception(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_ec2.instances.filter.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])
    with pytest.raises(AwsError):
        start_instance('foo-bar')

@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_stop_instance_resource_exception(mock_boto3):
    mock_boto3.resource.side_effect = ResourceNotExistsError(mock.Mock(), 'error', '')
    with pytest.raises(AwsError):
        stop_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_stop_instance_api_exception(mock_boto3):
    mock_boto3.resource.side_effect = UnknownAPIVersionError(mock.Mock(), 'error', '')
    with pytest.raises(AwsError):
        stop_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_stop_instance_exception(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_instance = mock.Mock()
    mock_ec2.instances.filter.return_value = mock_instance
    mock_instance.stop.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])
    with pytest.raises(AwsError):
        stop_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_stop_instance_filter_exception(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_ec2.instances.filter.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])
    with pytest.raises(AwsError):
        stop_instance('foo-bar')



@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_term_instance_resource_exception(mock_boto3):
    mock_boto3.resource.side_effect = ResourceNotExistsError(mock.Mock(), 'error', '')
    with pytest.raises(AwsError):
        terminate_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_term_instance_api_exception(mock_boto3):
    mock_boto3.resource.side_effect = UnknownAPIVersionError(mock.Mock(), 'error', '')
    with pytest.raises(AwsError):
        terminate_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_term_instance_exception(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_instance = mock.Mock()
    mock_ec2.instances.filter.return_value = mock_instance
    mock_instance.terminate.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])
    with pytest.raises(AwsError):
        terminate_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_term_instance_filter_exception(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.resource.return_value = mock_ec2
    mock_ec2.instances.filter.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])

    with pytest.raises(AwsError):
        terminate_instance('foo-bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_ec2_describe_instance(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.client.return_value = mock_ec2

    ec2_describe_instance('foo')
    mock_ec2.describe_instances.assert_called_once_with(InstanceIds=['foo'])


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_ec2_describe_instance_aws_error_on_client_error(mock_boto3):
    mock_ec2 = mock.Mock()
    mock_boto3.client.return_value = mock_ec2
    mock_ec2.describe_instances.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])

    with pytest.raises(AwsError):
        ec2_describe_instance('foo')


@mock.patch('twindb_infrastructure.providers.aws.ec2_describe_instance')
def test_get_instance_state(mock_describe):
    get_instance_state('foo')
    mock_describe.assert_called_once_with('foo')


@mock.patch('twindb_infrastructure.providers.aws.ec2_describe_instance')
def test_get_instance_state_aws_exception(mock_describe):
    mock_describe.side_effect = AwsError('AwsError exception')
    with pytest.raises(AwsError):
        get_instance_state('foo')


@mock.patch('twindb_infrastructure.providers.aws.ec2_describe_instance')
def test_get_instance_private_ip(mock_describe):
    get_instance_private_ip('foo')
    mock_describe.assert_called_once_with('foo')


@mock.patch('twindb_infrastructure.providers.aws.ec2_describe_instance')
def test_get_instance_private_ip_aws_exception(mock_describe):
    mock_describe.side_effect = AwsError('AwsError exception')
    with pytest.raises(AwsError):
        get_instance_private_ip('foo')


@mock.patch('twindb_infrastructure.providers.aws.ec2_describe_instance')
def test_get_instance_public_ip(mock_describe):
    get_instance_public_ip('foo')
    mock_describe.assert_called_once_with('foo')


@mock.patch('twindb_infrastructure.providers.aws.ec2_describe_instance')
def test_get_instance_public_ip_aws_exception(mock_describe):
    mock_describe.side_effect = AwsError('AwsError exception')
    with pytest.raises(AwsError):
        get_instance_public_ip('foo')


@pytest.mark.parametrize('response, expected_code', [
    (
        {
            'ResponseMetadata': {
                'HTTPStatusCode': 200
            }
        },
        True
    ),
    (
        {
            'ResponseMetadata': {
                'HTTPStatusCode': 404
            }
        },
        False
    )
])
@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_add_name_tag(mock_boto3, response, expected_code):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    mock_client.create_tags.return_value = response
    assert add_name_tag('foo', 'bar') == expected_code
    mock_client.create_tags.assert_called_once_with(
        Resources=['foo'],
        Tags=[
            {
                'Key': 'Name',
                'Value': 'bar'
            }
        ]
    )
    mock_boto3.client.assert_called_once_with('ec2')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_add_name_tag_exception(mock_boto3):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    mock_boto3.client.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])
    with pytest.raises(AwsError):
        add_name_tag('foo', 'bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_add_name_tag_value_exception(mock_boto3):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    mock_client.create_tags.side_effect = ValueError

    with pytest.raises(AwsError):
        add_name_tag('foo', 'bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_associate_address(mock_boto3):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    associate_address('foo', 'bar', 'bah')
    mock_boto3.client.assert_called_once_with('ec2')
    mock_client.associate_address.assert_called_once()


def test_associate_address_no_params():
    assert not associate_address('foo')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_associate_address_ec2_exception(mock_boto3):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    mock_boto3.client.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])

    with pytest.raises(AwsError):
        associate_address('foo', 'bar', 'bah')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_associate_address_address_exception(mock_boto3):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    mock_client.associate_address.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])

    with pytest.raises(AwsError):
        associate_address('foo', 'bar', 'bah')





from subprocess import call
import time
import boto3
from boto3.exceptions import ResourceNotExistsError, UnknownAPIVersionError
from botocore.exceptions import ClientError

from twindb_infrastructure import log
from twindb_infrastructure.providers.common import wait_sshd

AWS_REGIONS = [
    'us-east-1',
    'us-east-2',
    'us-west-1',
    'us-west-2',
    'ap-northeast-2',
    'ap-southeast-1',
    'ap-southeast-2',
    'ap-northeast-1',
    'eu-central-1',
    'eu-west-1'
]


class AwsError(Exception):
    """General error in AWS"""


def ec2_describe_instance(instance_id):
    """
    Describe EC2 instance by instance id and return an object with its
    properties

    :param instance_id: Instance id
    :type instance_id: str
    :raise: AwsError
    :return: object with instance properties
    :rtype: dict
    """
    try:
        client = boto3.client('ec2')
        return client.describe_instances(InstanceIds=[instance_id])
    except ClientError as err:
        raise AwsError(err)
    except ValueError as err:
        raise AwsError('Internal bug in boto3: %s', err)


def get_instance_state(instance_id):
    """
    Get state of instance by instance id

    :param instance_id: Instance id
    :type instance_id: str
    :raise: AwsError
    :return: instance state
    :rtype str
    """
    try:
        response = ec2_describe_instance(instance_id)
        state = response["Reservations"][0]["Instances"][0]["State"]["Name"]
        log.debug("Instance: %s, State: %s" % (instance_id, state))
        return state
    except AwsError as err:
        raise err


def get_instance_private_ip(instance_id):
    """
    Get private IP of instance by instance id

    :param instance_id: Instance id
    :type instance_id: str
    :raise: AwsError
    :return: private ip address of instance
    :rtype str
    """
    try:
        response = ec2_describe_instance(instance_id)
        return response["Reservations"][0]["Instances"][0]["PrivateIpAddress"]
    except AwsError as err:
        raise err


def get_instance_public_ip(instance_id):
    """
    Get public IP of instance by instance id

    :param instance_id: Instance id
    :type instance_id: str
    :raise: AwsError
    :return: public ip address of instance
    :rtype str
    """
    try:
        response = ec2_describe_instance(instance_id)
        return response["Reservations"][0]["Instances"][0]["PublicIpAddress"]
    except AwsError as err:
        raise err


def launch_ec2_instance(instance_profile, region=AWS_REGIONS[0],
                        private_key_file=None):
    """
    Launch instance

    :param instance_profile: Instance profile
    :param private_key_file: Private key file
    :param region: Region name
    :type instance_profile: dict
    :type private_key_file: str
    :type region: str
    :raise AwsError:
    :return: Instance id
    :rtype: str
    """
    try:
        client = boto3.client('ec2', region_name=region)
    except ClientError as err:
        raise AwsError(err)

    client_args = {
        'ImageId': instance_profile['ImageId'],
        'InstanceType': instance_profile['InstanceType'],
        'KeyName': instance_profile['KeyName'],
        'SubnetId': instance_profile["SubnetId"],
    }
    security_group_ids = list(instance_profile['SecurityGroupIds'])

    client_args['SecurityGroupIds'] = security_group_ids

    if instance_profile.get('MinCount'):
        client_args['MinCount'] = instance_profile["MinCount"]
    else:
        client_args['MinCount'] = 1

    if instance_profile.get('MaxCount'):
        client_args['MaxCount'] = instance_profile["MaxCount"]
    else:
        client_args['MaxCount'] = 1

    if instance_profile.get('AvailabilityZone'):
        client_args['Placement'] = {
            'AvailabilityZone': instance_profile['AvailabilityZone']
        }

    if instance_profile.get('EbsOptimized', False):
        client_args['EbsOptimized'] = True

    device_mappings = [
        {
            "DeviceName": "/dev/sda1",
            "Ebs": {
                "VolumeSize": instance_profile['RootVolumeSize'],
                "DeleteOnTermination": True,
                "VolumeType": "gp2"
            }
        }
    ]

    try:
        for volume in instance_profile['BlockDeviceMappings']:
            ebs = {
                "VolumeSize": volume['VolumeSize'],
                "DeleteOnTermination": True,
                "VolumeType": volume['VolumeType']
            }
            if 'Iops' in volume:
                ebs['Iops'] = volume['Iops']

            device_mappings.append({
                "DeviceName": volume['DeviceName'],
                "Ebs": ebs
            })
    except KeyError:
        pass

    client_args['BlockDeviceMappings'] = device_mappings
    try:
        response = client.run_instances(**client_args)
    except ClientError as err:
        raise AwsError(err)

    try:
        instance_id = response["Instances"][0]["InstanceId"]

        # Wait till instance is running
        timeout = 300
        log.info("Waiting till the instance starts")
        while get_instance_state(instance_id) != "running":
            log.info("Waiting %d more seconds" % timeout)

            time.sleep(3)
            timeout -= 3
            if timeout == 0:
                log.error("Failed to start instance. Timeout expired")
                return None

        if "Name" in instance_profile:
            try:
                add_name_tag(instance_id, instance_profile["Name"])
            except AwsError as err:
                raise err

        if "PublicIP" in instance_profile:
            public_ip = instance_profile['PublicIP']
            try:
                associate_address(instance_id, public_ip=public_ip)
            except AwsError as err:
                raise err

        # Wait will sshd is up
        try:
            ip = get_instance_public_ip(instance_id)
        except AwsError as err:
            raise err
        username = instance_profile["UserName"]
        if not wait_sshd(ip, private_key_file, username):
            return None

        if 'BlockDeviceMappings' in instance_profile:
            mount_volumes(ip, private_key_file, username,
                          volumes=instance_profile['BlockDeviceMappings'])

        if 'SourceDestCheck' in instance_profile:
            client.modify_instance_attribute(
                InstanceId=instance_id,
                SourceDestCheck={
                    'Value': instance_profile['SourceDestCheck']
                }
            )
        return instance_id
    except AwsError as err:
        raise AwsError(err)


def add_name_tag(instance_id, name):
    """
    Added tag to instance

    :param instance_id: Instance id
    :param name: Tag for instance
    :type instance_id: str
    :type name: str
    :raise: AwsError
    :return: Result of setting tag
    :rtype: bool
    """
    try:
        client = boto3.client('ec2')
        response = client.create_tags(
            Resources=[instance_id],
            Tags=[
                {
                    'Key': 'Name',
                    'Value': name
                }
            ]
        )
        return response['ResponseMetadata']['HTTPStatusCode'] == 200
    except ClientError as err:
        raise AwsError(err)
    except ValueError as err:
        raise AwsError('Internal bug in boto3: %s', err)


def associate_address(instance_id, public_ip=None, private_ip=None):
    """
    Associates an Elastic IP address with an instance

    :param instance_id: Instance id
    :param public_ip: Public ip
    :param private_ip: Private ip
    :type instance_id: str
    :type public_ip: str
    :type private_ip: str
    :raise: AwsError
    """

    kwargs = {
        'InstanceId': instance_id
    }
    if public_ip or private_ip:
        try:
            client = boto3.client('ec2')

            if public_ip:
                kwargs['PublicIp'] = public_ip

            if private_ip:
                kwargs['PrivateIpAddress'] = private_ip

            client.associate_address(**kwargs)

        except ClientError as err:
            raise AwsError(err)
        except ValueError as err:
            raise AwsError('Internal bug in boto3: %s', err)
    else:
        return None


def mount_volumes(ip, key_file, username, volumes=None):
    for volume in volumes:
        log.info("Mounting %s" % volume['MountPoint'])
        call(["ssh", "-t", "-o", "StrictHostKeyChecking=no",
              "-o", "PasswordAuthentication=no",
              "-i", key_file, "-l", username, ip,
              "sudo mkfs.xfs %s" % volume['DeviceName']])

        call(["ssh", "-t", "-o", "StrictHostKeyChecking=no",
              "-o", "PasswordAuthentication=no",
              "-i", key_file, "-l", username, ip,
              "sudo mkdir -p %s; sudo mount %s %s"
              % (volume['MountPoint'],
                 volume['DeviceName'],
                 volume['MountPoint'])])


def start_instance(instance_id):
    """
    Start Amazon instance

    :param instance_id: id of instance for run
    :type instance_id: str
    """
    try:
        ec2 = boto3.resource('ec2')
    except ResourceNotExistsError as err:
        raise AwsError(err)
    except UnknownAPIVersionError as err:
        raise AwsError(err)
    try:
        instance = ec2.instances.filter(InstanceIds=[instance_id])
        instance.start()
    except ClientError as err:
        raise AwsError(err)


def terminate_instance(instance_id):
    """
    Terminate Amazon instance

    :param instance_id: id of instance for terminate
    :type instance_id: str
    """
    try:
        ec2 = boto3.resource('ec2')
    except ResourceNotExistsError as err:
        raise AwsError(err)
    except UnknownAPIVersionError as err:
        raise AwsError(err)
    try:
        instance = ec2.instances.filter(InstanceIds=[instance_id])
        instance.terminate()
    except ClientError as err:
        raise AwsError(err)


def stop_instance(instance_id):
    """
    Stop Amazon instance

    :param instance_id: id of instance for stop
    :type instance_id: str
    """
    try:
        ec2 = boto3.resource('ec2')
    except ResourceNotExistsError as err:
        raise AwsError(err)
    except UnknownAPIVersionError as err:
        raise AwsError(err)
    try:
        instance = ec2.instances.filter(InstanceIds=[instance_id])
        instance.stop()
    except ClientError as err:
        raise AwsError(err)



import mock
import pytest
from botocore.exceptions import ClientError

from twindb_infrastructure.providers.aws import launch_ec2_instance, AwsError


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_launch_ec2_instance_client_aws_exception(mock_boto3):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    mock_boto3.client.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])

    with pytest.raises(AwsError):
        launch_ec2_instance({}, 'bar')


@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_launch_ec2_instance_run_instances_exception(mock_boto3):
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client
    instance_profile = {
        'ImageId': '',
        'InstanceType': '',
        'KeyName': '',
        'SubnetId': '',
        'SecurityGroupIds': [],
        'RootVolumeSize': 0,
        'MinCount': 1,
        'MaxCount': 1,
        'BlockDeviceMappings': [
            {
                'DeviceName': 'Such name',
                'VolumeSize': 0,
                'VolumeType': 'such type'
            }
        ],
    }
    mock_client.run_instances.side_effect = ClientError({'Error': {'Code': '404', 'Message': 'Not Found'}}, [])

    with pytest.raises(AwsError):
        launch_ec2_instance(instance_profile)


@mock.patch('twindb_infrastructure.providers.aws.wait_sshd')
@mock.patch('twindb_infrastructure.providers.aws.get_instance_public_ip')
@mock.patch('twindb_infrastructure.providers.aws.get_instance_state')
@mock.patch('twindb_infrastructure.providers.aws.boto3')
def test_calls_with_srcdst_check(mock_boto3,
                                 mock_get_instance_state,
                                 mock_get_instance_public_ip,
                                 mock_wait_sshd):

    mock_wait_sshd.return_value = True
    mock_get_instance_public_ip.return_value = 'some ip'
    mock_get_instance_state.return_value = 'running'
    mock_client = mock.Mock()
    mock_boto3.client.return_value = mock_client

    response = {
        'Instances': [
            {
                'InstanceId': 'foo'
            }
        ]
    }
    mock_client.run_instances.return_value = response
    instance_profile = {
        'ImageId': '',
        'InstanceType': '',
        'KeyName': '',
        'SubnetId': '',
        'SecurityGroupIds': [],
        'RootVolumeSize': 0,
        'MinCount': 1,
        'MaxCount': 1,
        'UserName': 'centos',
        'SourceDestCheck': False,
    }
    instance_id = launch_ec2_instance(instance_profile)

    mock_client.modify_instance_attribute.assert_called_once_with(
        SourceDestCheck={
            'Value': False
        },
        InstanceId=instance_id
    )






import pytest

from twindb_infrastructure.config.config import ConfigException
from twindb_infrastructure.util import parse_config, domainname


def test_parse_config_raises_config_exception():
    with pytest.raises(ConfigException):
        parse_config('/foo/bar')


@pytest.mark.parametrize('host, domain', [
    (
        'www.google.com',
        'google.com'
    ),
    (
        'www.google.com.',
        'google.com.'
    )
])
def test_domainname(host, domain):
    assert domainname(host) == domain



from click.testing import CliRunner

from twindb_infrastructure import twindb_galera


def test_main():
    runner = CliRunner()
    result = runner.invoke(twindb_galera.main)
    assert result.exit_code == 0



# -*- coding: utf-8 -*-
"""
TwinDB Amazon infrastructure
"""
from subprocess import CalledProcessError

import boto3
import time
from botocore.exceptions import ClientError
import click
import json
from twindb_infrastructure import setup_logging, __version__
from twindb_infrastructure import log
from twindb_infrastructure.config.config import TWINDB_INFRA_CONFIG, \
    ConfigException
from twindb_infrastructure.providers.aws import AWS_REGIONS, \
    launch_ec2_instance, get_instance_private_ip
from twindb_infrastructure.switchover import change_names_to, \
    log_remaining_sessions, stop_proxy, eth1_present, start_proxy, \
    restart_proxy, server_ready
from twindb_infrastructure.tagset import TagSet
from twindb_infrastructure.util import printf, parse_config


class TwinDBInfraException(Exception):
    pass


CONFIG = None


@click.group(invoke_without_command=True)
@click.option('--config', default=TWINDB_INFRA_CONFIG,
              help='Config file',
              show_default=True,
              type=click.Path())
@click.option('--debug', is_flag=True, default=False,
              help='Print debug messages')
@click.option('--version', help='Show tool version and exit.', is_flag=True,
              default=False)
@click.pass_context
def main(ctx, config, debug, version):
    """
    Console script to work with TwinDB Amazon Infrastructure
    """
    global CONFIG

    if not ctx.invoked_subcommand:
        if version:
            print(__version__)
        else:
            print(ctx.get_help())
        return 0

    setup_logging(log, debug=debug)
    log.debug('Using config %s' % config)
    try:
        CONFIG = parse_config(config)
    except ConfigException as err:  # pragma: no cover
        log.error(err)
        exit(-1)


@main.command()
@click.option('--tags', is_flag=True, help='Show instance tags')
@click.option('--verbose', is_flag=True,
              help='Show more details about the instance')
@click.option('--region', type=click.Choice(AWS_REGIONS),
              help='AWS region name')
@click.argument('tags-filter', nargs=-1)
def show(tags, verbose, region, tags_filter):
    """List TwinDB servers"""

    if region:
        region_name = region
    else:
        region_name = CONFIG.aws.aws_default_region

    client = boto3.client('ec2', region_name=region_name)
    response = client.describe_instances()
    log.debug('response = %r' % response)
    for reservation in response['Reservations']:
        for instance in reservation['Instances']:
            if tags_filter \
                    and \
                    not TagSet(tags_filter).issubset(TagSet(instance['Tags'])):
                continue

            printf('%s' % instance['InstanceId'])

            if tags:
                printf(': Tags: ')
                try:
                    printf('%s', TagSet(instance['Tags']))
                except KeyError:
                    printf("None")
                    pass

            if verbose:
                if tags:
                    printf(",")
                printf(' State: %s,' % instance['State']['Name'])
                try:
                    printf(' PublicIP: %s,' % instance['PublicIpAddress'])
                except KeyError:
                    pass
                try:
                    printf(' PrivateIP: %s,' % instance['PrivateIpAddress'])
                except KeyError:
                    pass
                printf(' AvailabilityZone: %s'
                       % instance['Placement']['AvailabilityZone'])

            printf("\n")


@main.command()
@click.argument('instance-id')
def terminate(instance_id):
    """Terminate Amazon instance"""

    try:
        ec2 = boto3.resource('ec2')
        ec2.instances.filter(InstanceIds=[instance_id]).terminate()

    except ClientError as err:
        log.error(err)


@main.command()
@click.argument('instance-id')
def stop(instance_id):
    """Stop Amazon instance"""

    try:
        ec2 = boto3.resource('ec2')
        ec2.instances.filter(InstanceIds=[instance_id]).stop()

    except ClientError as err:
        log.error(err)


@main.command()
@click.argument('instance-id')
def start(instance_id):
    """Start Amazon instance"""

    try:
        ec2 = boto3.resource('ec2')
        ec2.instances.filter(InstanceIds=[instance_id]).start()

    except ClientError as err:
        log.error(err)


@main.command()
@click.argument('template')
def launch(template):
    """Launch new Amazon instance"""

    log.info("Starting instance")
    with open(template) as fp:
        instance_profile = json.loads(fp.read())

        private_key_file = CONFIG.ssh.private_key_file
        instance_id = \
            launch_ec2_instance(instance_profile,
                                private_key_file=private_key_file)

        if not instance_id:
            log.error("Failed to launch EC2 instance")
            exit(-1)

        log.info("Launched instance %s" % instance_id)

        ip = get_instance_private_ip(instance_id)
        log.info('Instance %s on %s' % (instance_id, ip))


@main.command()
@click.argument('proxy_a')
@click.argument('proxy_b')
@click.argument('vip')
@click.option('--dns', multiple=True,
              help='Update given DNS names.')
@click.option('--mysql-user', help='MySQL user to connect to proxies.',
              default='root', show_default=True)
@click.option('--mysql-password', help='Password for MySQL user.',
              default='', show_default=True)
@click.option('--mysql-port', help='TCP port that open for MySQL connections.',
              default=3306, show_default=True)
def switch_proxy(proxy_a, proxy_b, vip, dns,
                 mysql_user, mysql_password, mysql_port):
    """Switch active ProxySQL from proxy_a to proxy_b.

    proxy_a is a currently active proxy.

    proxy_b is to be active proxy.

    VIP is virtual IP.
    """
    for host in [proxy_a, proxy_b, vip]:
        if not server_ready(host,
                            user=mysql_user,
                            password=mysql_password,
                            port=mysql_port):
            log.error('Server %s must be up and running to do switchover',
                      host)
            exit(1)
    if proxy_a == proxy_b:
        log.error('Proxy A and Proxy B cannot be same')
        exit(1)

    if vip in [proxy_a, proxy_b]:
        log.error('VIP address cannot be equal to proxy A or proxy B')
        exit(1)

    log.info("Switching active ProxySQL from %s to %s", proxy_a, proxy_b)
    log.debug('DNS names: %s', ', '.join(dns))

    if not eth1_present(proxy_a):
        log.error('It looks like %s is not active proxy', proxy_a)
        exit(1)

    if eth1_present(proxy_b):
        log.error('Interface eth1 is expected on %s, but found on %s. '
                  'Exiting', proxy_a, proxy_b)
        exit(1)

    # Step 1. Restart ProxyB
    try:
        restart_proxy(proxy_b)
        pass
    except CalledProcessError as err:
        log.error(err)
        exit(1)

    # Step 2, 3. Change DNS so dns points to Proxy B Private IP
    change_names_to(dns, proxy_b)

    # Step 4. Wait TTL * 2 time
    wait_time = 30 * 60
    timeout = time.time() + wait_time
    n_conn = 0
    while time.time() < timeout:
        # Step 5. Check if any MySQL users are connected to Proxy A
        #   (and log if any)
        n_conn = log_remaining_sessions(proxy_a,
                                        mysql_user,
                                        mysql_password,
                                        mysql_port)
        if not n_conn:
            log.info('All sessions disconnected from %s', proxy_a)
            break
        log.info('There are still %d open sessions on %s', n_conn, proxy_a)
        time.sleep(3)
    if n_conn > 0:
        log.warning('Will kill existing sessions anyway')
    # 6. Stop Proxy A
    stop_proxy(proxy_a)

    # 7. Wait until eth1 shows up on Proxy B. If not - log error and stop.
    wait_time = 60
    timeout = time.time() + wait_time
    while time.time() < timeout:
        if eth1_present(proxy_b) and server_ready(vip,
                                                  user=mysql_user,
                                                  password=mysql_password,
                                                  port=mysql_port):
            break
        time.sleep(3)

    if not eth1_present(proxy_b):
        log.error('Keepalived failed to move VIP to %s', proxy_b)
        exit(1)

    # 8. Start Proxy A
    start_proxy(proxy_a)

    # 9. Change DNS so dns points to points to VIP
    # 10. Change DNS so dns points to points to VIP
    change_names_to(dns, vip)
    log.info('Successfully done.')



class Tag(object):
    def __init__(self, tag=None, key=None, value=None):
        """
        Init Tag instance from a dictionary

        :param tag: dictionary with a tag definition like
            {u'Value': 'nat', u'Key': 'Name'}
        :param key: string with a tag key
        :param value: string with a tag value
        """
        try:
            self.key = tag['Key']
            self.value = tag['Value']
        except TypeError:
            if isinstance(tag, basestring):
                split_tag = tag.split('=')
                self.key = split_tag[0]
                self.value = split_tag[1]
            elif key and value:
                self.key = key
                self.value = value
            else:
                raise

    def __eq__(self, other):
        return self.key == other.key and self.value == other.value


class TagSet(object):

    def __init__(self, tags=None):
        """
        Init tagset instance from list of tags

        :param tags: list of dictionaries with tags. For example,
            [{u'Value': 'nat', u'Key': 'Name'}]
        """
        self.tags = set()
        for tag_def in tags:
            tag = Tag(tag_def)
            self.tags.add(tag)

    def find(self, tag):
        """
        FInd in a set a tag equal to the given tag
        :param tag: Tag() instance
        :return: Tag() or None if not found
        """
        for t in self.tags:
            if t == tag:
                return t

        return None

    def issubset(self, other):
        """
        Test whether every element in the set is in other.

        :param other: instance of TagSet()
        :return: True or False
        """
        for tag in self.tags:
            if not other.find(tag):
                return False

        return True

    def __repr__(self):
        comma = ''
        space = ''
        result = ''
        for tag in self.tags:
            result += "{comma}{space}{key}={value}".format(
                key=tag.key,
                value=tag.value,
                comma=comma,
                space=space)
            comma = ','
            space = ' '
        return result

    def __eq__(self, other):
        return self.issubset(other) and other.issubset(self)



from ConfigParser import ConfigParser
from twindb_infrastructure.config import TWINDB_INFRA_CONFIG
from twindb_infrastructure.config.aws_credentials import AwsCredentials
from twindb_infrastructure.config.cloudflare_credentials import \
    CloudFlareCredentials
from twindb_infrastructure.config.credentials import CredentialsException
from twindb_infrastructure.config.ssh_credentials import SshCredentials


class ConfigException(Exception):
    pass


class Config(object):
    config = None
    config_path = None
    aws = None
    ssh = None
    cloudflare = None

    def __init__(self, config_path=TWINDB_INFRA_CONFIG):
        self.config_path = config_path
        self.config = ConfigParser()
        self.config.read(config_path)
        try:
            self.aws = AwsCredentials(config_path=self.config_path)
            self.ssh = SshCredentials(config_path=self.config_path)
            self.cloudflare = CloudFlareCredentials(
                config_path=self.config_path)

        except CredentialsException as err:
            raise ConfigException(err)



from textwrap import dedent

import mock

from twindb_infrastructure.loader import Loader


@mock.patch('twindb_infrastructure.loader.requests')
def test_loader(mock_requests):
    mock_response = mock.Mock()
    mock_response.content = dedent(
        """
        <html>
        <title>foo</title>
        <body>bar</body>
        </html>
        """
    )
    mock_requests.get.return_value = mock_response
    loader = Loader('xxx')
    assert loader.title == 'foo'
    assert loader.body == 'bar'


@mock.patch('twindb_infrastructure.loader.requests')
def test_loader_no_redirects(mock_requests):
    loader = Loader('xxx')
    assert loader.title is None
    mock_requests.get.assert_called_once_with(
        'xxx', allow_redirects=False, timeout=10
    )


@mock.patch('twindb_infrastructure.loader.requests')
def test_loader_with_headers(mock_requests):
    loader = Loader('xxx', host='foo.bar', protocol='aaa')
    assert loader.title is None
    mock_requests.get.assert_called_once_with(
        'xxx', allow_redirects=False, timeout=10,
        headers={
            'Host': 'foo.bar',
            'X-Forwarded-Proto': 'aaa',
        }
    )



# -*- coding: utf-8 -*-
import logging
import logging.handlers

from logutils.colorize import ColorizingStreamHandler

__author__ = 'TwinDB Development Team'
__email__ = 'dev@twindb.com'
__version__ = '1.4.1'

log = logging.getLogger(__name__)


def setup_logging(logger, debug=False):

    fmt_str = "%(asctime)s: %(levelname)s:" \
              " %(module)s.%(funcName)s():%(lineno)d: %(message)s"
    ColorizingStreamHandler.level_map = {
        logging.DEBUG: (None, 'blue', False),
        logging.INFO: (None, 'green', False),
        logging.WARNING: (None, 'yellow', False),
        logging.ERROR: (None, 'red', False),
        logging.CRITICAL: ('red', 'white', True),
    }
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter(fmt_str))
    logger.handlers = []
    logger.addHandler(console_handler)
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)



import click
from subprocess import Popen
from twindb_infrastructure import setup_logging, __version__
from twindb_infrastructure import log
from twindb_infrastructure.config import TWINDB_INFRA_CONFIG
from twindb_infrastructure.config.config import ConfigException
from twindb_infrastructure.util import parse_config, stop_chef_client, \
    stop_galera, remote_rmdir, remote_restore, \
    bootstrap_first_node, bootstrap_next_node


CONFIG = None


@click.group(invoke_without_command=True)
@click.option('--config', default=TWINDB_INFRA_CONFIG,
              help='Config file',
              show_default=True,
              type=click.Path())
@click.option('--debug', is_flag=True, default=False,
              help='Print debug messages')
@click.option('--version', help='Show tool version and exit.', is_flag=True,
              default=False)
@click.pass_context
def main(ctx, config, debug, version):
    """
    Console script to work with Galera
    """
    global CONFIG
    if not ctx.invoked_subcommand:
        if version:
            print(__version__)
        else:
            print(ctx.get_help())
        return 0

    setup_logging(log, debug=debug)
    log.debug('Using config %s' % config)
    try:
        CONFIG = parse_config(config)
    except ConfigException as err:  # pragma: no cover
        log.error(err)
        exit(-1)


@main.command()
@click.argument('backup_copy', required=False)
@click.option('--datadir', help='Directory where to restore the backup copy',
              default='/var/lib/mysql', show_default=True)
@click.option('--founder', help='IP address of a founder node',
              show_default=True)
@click.option('--node', '-n', multiple=True,
              help='IP addresses on cluster nodes. '
                   'Multiple options are allowed')
def cluster(backup_copy, datadir, founder, node):
    """Bootstrap xtradb cluster"""
    if backup_copy:
        all_nodes = list(node) + [founder]
        stop_chef_client(all_nodes)
        stop_galera(all_nodes)
        remote_rmdir(datadir, all_nodes)

        remote_restore(founder, backup_copy, datadir)
        bootstrap_first_node(founder, datadir)
        bootstrap_next_node(node[0], datadir)
        # for n in node:
        #    bootstrap_next_node(n, datadir)
        # start_chef_client(all_nodes)
        pass
    else:
        log.info('No backup copy specified. Choose one from below:')
        proc = Popen(['twindb-backup', 'ls'])
        proc.communicate()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


def parse_requirements(path):
    """
    Parses requirements from a file given by path and return a list consumable by setuptools.

    :param path: Path to requirements file
    :type path: str
    :return: List of requirements.
    :rtype: list
    """
    with open(path) as fp:
        reqs = fp.read().strip().split('\n')

    return [x for x in reqs if x and not x.strip().startswith('#')]


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = parse_requirements('requirements.txt')

test_requirements = parse_requirements('requirements_dev.txt')

setup(
    name='twindb-infrastructure',
    version='1.4.1',
    description="TwinDB Infrastructure is a collection of everything"
                " to manage TwinDB infrastructure",
    long_description=readme + '\n\n' + history,
    author="TwinDB Development Team",
    author_email='dev@twindb.com',
    url='https://github.com/twindb/twindb_infrastructure',
    packages=find_packages(),
    package_dir={'twindb_infrastructure':
                 'twindb_infrastructure'},
    entry_points={
        'console_scripts': [
            'twindb-aws=twindb_infrastructure.twindb_aws:main',
            'twindb-chef=twindb_infrastructure.twindb_chef:main',
            'twindb-galera=twindb_infrastructure.twindb_galera:main',
            'twindb-monitoring=twindb_infrastructure.monitoring:main',
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='twindb_infrastructure',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



import pytest
from twindb_infrastructure.tagset import TagSet, Tag


@pytest.fixture
def sample_tag_definition():
    return {
        'Key': 'Name',
        'Value': 'foo'
    }


def test_tag(sample_tag_definition):
    tag = Tag(sample_tag_definition)
    assert tag.key == sample_tag_definition['Key']
    assert tag.value == sample_tag_definition['Value']


def test_tags_equal(sample_tag_definition):
    tag1 = Tag(sample_tag_definition)
    tag2 = Tag(sample_tag_definition)
    assert tag1 == tag2


def test_tagset_init(sample_tag_definition):
    tags = [sample_tag_definition]
    TagSet(tags)


def test_tagset_find(sample_tag_definition):
    tags = [sample_tag_definition]
    tagset = TagSet(tags)
    tag = Tag(sample_tag_definition)
    assert tagset.find(tag)


def test_tagset_issubset():
    big_set = TagSet([
        {
            'Key': 'tag1',
            'Value': 'value1'
        },
        {
            'Key': 'tag2',
            'Value': 'value2'
        },
        {
            'Key': 'tag3',
            'Value': 'value3'
        }
    ])

    small_set = TagSet([
        {
            'Key': 'tag1',
            'Value': 'value1'
        },
        {
            'Key': 'tag2',
            'Value': 'value2'
        }
    ])

    print(big_set)
    print(small_set)
    assert small_set.issubset(big_set)
    assert not big_set.issubset(small_set)


def test_tag_from_str():
    assert Tag('Name=aaa') == Tag(key='Name', value='aaa')


def test_tagset_from_tuple():
    assert TagSet((u'Name=aaa', u'Role=master')) == TagSet([
        {
            'Key': 'Name',
            'Value': 'aaa'
        },
        {
            'Key': 'Role',
            'Value': 'master'
        }
    ])


def test_tag_raises_error():
    with pytest.raises(TypeError):
        Tag(tag=None)


def test_find_returns_none():
    ts = TagSet(['Name=foo'])
    assert ts.find(Tag('Name=bar')) is None



"""Module with Loader() class"""
from HTMLParser import HTMLParser

import requests
import time


class Loader(object, HTMLParser):
    def __init__(self, url, timeout=10, host=None, protocol=None):
        HTMLParser.__init__(self)
        self._url = url
        self._timeout = timeout
        self.__response = None
        self._body = None
        self._title = None
        self.__current_tag = None
        self._host = host
        self._protocol = protocol
        self._load_time = None

    @property
    def body(self):
        return self._get_tag('body')

    @property
    def load_time(self):
        if self._load_time is None:
            start = time.time()
            self.load()
            self._load_time = time.time() - start

        return self._load_time

    @property
    def title(self):
        return self._get_tag('title')

    @property
    def url(self):
        return self._url

    def handle_starttag(self, tag, attrs):
        self.__current_tag = tag

    def handle_endtag(self, tag):
        self.__current_tag = None

    def handle_data(self, data):
        setattr(self, '_%s' % self.__current_tag, data)

    def load(self):
        return self._response

    def reset(self):
        super(Loader, self).reset()
        self.__response = None
        self._body = None
        self._title = None
        self._load_time = None

    @property
    def _response(self):
        if self.__response is None:
            kwargs = {
                'timeout': self._timeout,
                'allow_redirects': False,
            }
            headers = {}
            if self._host:
                headers['Host'] = self._host
            if self._protocol:
                headers['X-Forwarded-Proto'] = self._protocol
            if headers:
                kwargs['headers'] = headers

            resp = requests.get(self._url, **kwargs)
            resp.raise_for_status()
            self.__response = resp.content

        return self.__response

    def _get_tag(self, tag):
        if getattr(self, '_%s' % tag) is None:
            self.feed(self._response)

        return getattr(self, '_%s' % tag)



import ConfigParser
import os
from twindb_infrastructure.config import TWINDB_INFRA_CONFIG


class CredentialsException(Exception):
    """
    Exception for Credentials class errors
    """


class Credentials(object):
    """
    Class to work with TwinDB Infrastructure config file
    """
    config_path = None
    """TwinDB Intrastructure Config"""

    def __init__(self, config_path=TWINDB_INFRA_CONFIG):
        """

        :type config_path: str
        """
        if not os.path.exists(config_path):
            raise CredentialsException("Credentials config file %s not found"
                                       % config_path)
        self.config_path = config_path
        try:
            self.config = ConfigParser.ConfigParser()
            self.config.read(self.config_path)

        except ConfigParser.Error as err:
            raise CredentialsException(err)

        if not self.config.sections():
            raise CredentialsException('No sections in config %s'
                                       % self.config_path)
